/*
 * Copyright 2002 Sun Microsystems, Inc.  All rights reserved.
 *
 * Use is subject to the GNU General Public License, Version 2,
 * June 1991, which is contained in the read-me file named
 * "README_GNU_GPL." This program is free software; you can
 * redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your
 * option) any later version. This library is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <netdb.h>

#include "lcdutils.h"
#include "lcd.h"

#define TIMEOUTSECS 30 

void timeout_handler(int sig) 
{
  lcd_unlock();
  system("/sbin/lcdstop");
  system("/etc/rc.d/init.d/lcd-showip");
}


int main(int argc, char **argv) 
{

  char *storeline1 = NULL, *storeline2 = NULL;
  char *line1, *line2;
  void *lcd;

  int 	sleeper=0;
  int 	dum=52;

  lcd_setlocale();
  system("/sbin/lcdstop");

  if ((lcd = lcd_open(O_RDWR)) == NULL) {
    printf(_("LCD is not present\n"));
    exit(1);
  }

  while (1) {
    lcd_wait_no_button(lcd);
    lcd_timeout_start(timeout_handler, TIMEOUTSECS);
    if ((dum = lcd_getbutton(lcd)) < 0) 
      continue; /* try again */

    printf(_("LCD Awake\n"));
    lcd_timeout_stop();
    system("/sbin/lcdstop");
    lcd_lock();
      
    sleeper = 0;
      
    lcd_getdisplay(lcd, &line1, &line2, NULL); /* read in the display */

    /* free if necessary */
    if (storeline1)
      free(storeline1);
    if (storeline2)
      free(storeline2);

    storeline1 = line1 ? strdup(line1) : NULL;
    storeline2 = line2 ? strdup(line2) : NULL;
      
    if (dum == BUTTON_Power) {
      system("/etc/lcd.d/power_down -s");
      system("/etc/rc.d/init.d/lcd-showip");
      sleeper = 1;
    }

    /*  Press the Reset Password button */
    if (dum == 0xFC) {
      system("/etc/lcd.d/reset_password");
      sleeper = 1;
    }
    
    while (!sleeper) {
      lcd_reset(lcd);
      sleeper = system("/sbin/lcd-menu /etc/lcd.d");
      lcd_write(lcd, storeline1, storeline2);
      printf(_("LCD Sleeping\n"));
      system("/etc/rc.d/init.d/lcd-showip");
    }
    lcd_unlock();
  }

  lcd_close(lcd);
}

